!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! STRINGFAST: Module to compute CMB anisotropy spectra induced by cosmic
!   strings with specified tension, "wiggliness," rms transverse velocity,
!   and correlation length
!
! First described in the paper Foreman, Moss, and Scott (2011)
!  (arXiv:1106.4018)
!
! If you make use of StringFast, please cite
!     1. S. Foreman, A. Moss, and D. Scott (2011),
!          arXiv:1106.4018
!     2. L. Pogosian and T. Vachaspati, Phys. Rev. D60, 083504 (1999),
!          arXiv:astro-ph/9903361.
!
! Updates available at [website]
!
! Developed by Simon Foreman (sfore@stanford.edu)
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Permission to use, copy, modify and distribute without fee or royalty at
! any tier, this software and its documentation, for any purpose and
! without fee or royalty is hereby granted, provided that you agree to
! comply with the following copyright notice and statements, including the
! disclaimer, and that the same appear on ALL copies of the software and
! documentation, including modifications that you make for internal use or
! for distribution:
!
! Copyright 2011 by University of British Columbia.  All rights reserved.
!
! THIS SOFTWARE IS PROVIDED "AS IS", AND U.B.C. MAKES NO 
! REPRESENTATIONS OR WARRANTIES, EXPRESS OR IMPLIED.  
! BY WAY OF EXAMPLE, BUT NOT LIMITATION,
! U.B.C. MAKES NO REPRESENTATIONS OR WARRANTIES OF 
! MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR PURPOSE OR THAT 
! THE USE OF THE LICENSED SOFTWARE OR DOCUMENTATION WILL NOT INFRINGE 
! ANY THIRD PARTY PATENTS, COPYRIGHTS, TRADEMARKS OR OTHER RIGHTS.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! Description: Uses a grid (in parameter space) of pre-computed spectra
!  to interpolate spectra at arbitrary values of the string tension,
!  "wiggliness," velocity, and correlation length of a network of strings
!  described by the one-scale model for wiggly strings described in
!  Pogosian & Vachaspati, Phys. Rev. D60, 083504 (1999). CMBACT, the code
!  described in that paper, has been used to generate the pre-computed
!  curves, which have then been fit either with simple fitting functions
!  or cubic splines. The fit parameters and/or spline points are stored
!  in included parameter files (*.sfp), which are read in at runtime and
!  used by StringFast to interpolate the C_l curves at specified values
!  of the string parameters.
!
!  A flat FRW background cosmology is assumed, with WMAP7 parameters
!  from Larson et al., Astrophys. J. Suppl. 192, 16 (2011),
!  with the exception of the optical depth to reionization, which can be
!  specified by the user within the current limits of the grid.
!
! Usage:
!
!  ------------
!     SF_setuptt(svt), SF_setupbb(svt), SF_setupee(svt):
!
!       -set up variables necessary for calculations
!       -call at initialization
!
!       ARGUMENTS: svt: 1 = scalars
!                       2 = vectors
!                       3 = tensors
!
!       RETURN VALUE: none
!
!  ------------
!     SF_ttcalc(svt,gmu,od,alpha,vel,cl,ell),
!     SF_bbcalc(svt,gmu,od,alpha,vel,cl,ell),
!     SF_eecalc(svt,gmu,od,alpha,vel,cl,ell):
!
!       -calculate [l(l+1)]/2pi C_l (muK^2) for TT, BB, or EE mode resp.,
!          at specified l value, string parameter values, and
!          optical depth of background cosmology
!
!       ARGUMENTS: svt: 1 = scalars
!                       2 = vectors
!                       3 = tensors
!                  gmu: dimensionless string tension
!                  od: optical depth to reionization
!                           (between 0.085 and 0.095)
!                  alpha: wiggliness
!                           (between 1.0 and 1.3)
!                  vel: rms transverse velocity of strings
!                           (between 0.3 and 0.8)
!                  cl: comoving correlation length
!                           (between 0.15 and 0.50)
!                  ell: as in C_ell
!                           (between 2 and 2999)
!
!       RETURN VALUE: [l(l+1)]/2pi C_l, in muK^2
!
!  ------------
!     SF_finishtt(), SF_finishbb(), SF_finishee()
!
!       -deallocate memory allocated by corresponding SF_setup** calls
!
!       ARGUMENTS: none
!
!       RETURN VALUE: none
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

! Module with constaints and variable definitions (not for end-user
!  access)
module sf_vars

  ! Number of grid points in each parameter
  integer, parameter :: od_pts    = 3
  integer, parameter :: alpha_pts = 6
  integer, parameter :: vel_pts   = 11
  integer, parameter :: cl_pts    = 8
  ! Mix/max values of parameters in grid
  real(8), parameter :: grid_od_min     = 0.085d0
  real(8), parameter :: grid_od_max     = 0.095d0
  real(8), parameter :: grid_alpha_min  = 1.0d0
  real(8), parameter :: grid_alpha_max  = 1.5d0
  real(8), parameter :: grid_vel_min    = 0.30d0
  real(8), parameter :: grid_vel_max    = 0.80d0
  real(8), parameter :: grid_cl_min     = 0.15d0
  real(8), parameter :: grid_cl_max     = 0.50d0
  ! Number of fit parameters and/or spline points used to describe
  !  each type of spectrum
  integer, parameter :: tts_npars = 39
  integer, parameter ::   tts_splpts = 20
  integer, parameter :: ttv_npars = 48
  integer, parameter ::   ttv_splpts = 46
  integer, parameter ::   ttv_splpts_a = 17
  integer, parameter ::   ttv_splpts_b = 9
  integer, parameter ::   ttv_splpts_c = 11
  integer, parameter ::   ttv_splpts_d = 9
  integer, parameter :: ttt_npars = 40
  integer, parameter ::   ttt_splpts = 15
  integer, parameter :: bbv_npars = 35
  integer, parameter ::   bbv_splpts = 33
  integer, parameter :: bbt_npars = 35
  integer, parameter ::   bbt_splpts = 33
  integer, parameter ::   bbt_splpts_a = 16
  integer, parameter ::   bbt_splpts_b = 10
  integer, parameter :: ees_npars = 40
  integer, parameter ::   ees_splpts = 38
  integer, parameter :: eev_npars = 35
  integer, parameter ::   eev_splpts = 33
  integer, parameter :: eet_npars = 35
  integer, parameter ::   eet_splpts = 33
  integer, parameter ::   eet_splpts_a = 14
  integer, parameter ::   eet_splpts_b = 10
  ! Parameter file names
  character(LEN=30), parameter :: tts_par_file='params_tts.sfp'
  character(LEN=30), parameter :: ttv_par_file='params_ttv.sfp'
  character(LEN=30), parameter :: ttt_par_file='params_ttt.sfp'
  character(LEN=30), parameter :: bbv_par_file='params_bbv.sfp'
  character(LEN=30), parameter :: bbt_par_file='params_bbt.sfp'
  character(LEN=30), parameter :: ees_par_file='params_ees.sfp'
  character(LEN=30), parameter :: eev_par_file='params_eev.sfp'
  character(LEN=30), parameter :: eet_par_file='params_eet.sfp'
  
  ! Storage for quantities read and assembled from parameter files
  type ParSplines
    real(8) :: ods(1:od_pts)
    real(8) :: alphas(1:alpha_pts)
    real(8) :: vels(1:vel_pts)
    real(8) :: cls(1:cl_pts)
    real(8), allocatable :: sp(:,:,:,:,:)
    real(8), allocatable :: sp2(:,:,:,:,:)
  end type
  type (ParSplines), pointer :: tts_spls,ttv_spls,ttt_spls
  type (ParSplines), pointer :: bbv_spls,bbt_spls
  type (ParSplines), pointer :: ees_spls,eev_spls,eet_spls
  
  ! Flags to track which splines have been allocated
  logical :: bb_splined(2:3)=.false.
  logical :: tt_splined(1:3)=.false.
  logical :: ee_splined(1:3)=.false.
  logical :: te_splined(1:3)=.false.
  
  ! Tolerance for checking if parameters are the same as the last call
  real(8), parameter :: par_tol=1.0d-8
  
end module
!~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
!~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

! The main module, to use in any subroutine that calls StringFast
module stringfast

  ! Value of Gmu used for the pre-computed curves
  real(8), parameter :: gmu_grid = 6.4d-7
  ! Flag to print debugging messages
  logical, parameter :: sf_debug=.false.
  
contains
!~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
subroutine SF_setuptt(svt)
  ! Call subroutines that set up variables for TT calculations
  use sf_vars
  implicit none
  integer, intent(in) :: svt
  integer :: i,j,k,m
  real(8) :: tester,tester_x
  real(8) :: z(1:cl_pts),z2(1:cl_pts)
  
  ! Switch based on scalar, vector, or tensor modes: allocates splines
  !  and calls routine to set them up in memory
  select case (svt)
    case (1)  ! scalars
      allocate(tts_spls)
      allocate(tts_spls%sp(1:od_pts,1:alpha_pts,1:vel_pts,1:cl_pts,1:tts_npars))
      allocate(tts_spls%sp2(1:od_pts,1:alpha_pts,1:vel_pts,1:cl_pts,1:tts_npars))
      
      call setup_parsplines(tts_npars,tts_spls,tts_par_file)
      tt_splined(1)=.true.
      if (sf_debug) write(*,*) 'TT scalar splines set up'
    case (2)  ! vectors
      allocate(ttv_spls)
      allocate(ttv_spls%sp(1:od_pts,1:alpha_pts,1:vel_pts,1:cl_pts,1:ttv_npars))
      allocate(ttv_spls%sp2(1:od_pts,1:alpha_pts,1:vel_pts,1:cl_pts,1:ttv_npars))
      
      call setup_parsplines(ttv_npars,ttv_spls,ttv_par_file)
      tt_splined(2)=.true.
      if (sf_debug) write(*,*) 'TT vector splines set up'
    case (3)  ! tensors
      allocate(ttt_spls)
      allocate(ttt_spls%sp(1:od_pts,1:alpha_pts,1:vel_pts,1:cl_pts,1:ttt_npars))
      allocate(ttt_spls%sp2(1:od_pts,1:alpha_pts,1:vel_pts,1:cl_pts,1:ttt_npars))
      
      call setup_parsplines(ttt_npars,ttt_spls,ttt_par_file)
      tt_splined(3)=.true.
      if (sf_debug) write(*,*) 'TT tensor splines set up'
    case default
      call sf_fatalerror('Bad SF_setuptt call - argument must be 1,2,3')
  end select

end subroutine
!~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
subroutine SF_setupbb(svt)
  ! Call subroutines that set up variables for BB calculations
  use sf_vars
  implicit none
  integer, intent(in) :: svt
  integer :: i,j,k,m
  real(8) :: tester,tester_x
  real(8) :: z(1:cl_pts),z2(1:cl_pts)
  
  select case (svt)
    case (1)
      call sf_fatalerror('Bad SF_setupbb call - there are no BB scalar modes!')
    case (2)  ! vectors
      allocate(bbv_spls)
      allocate(bbv_spls%sp(1:od_pts,1:alpha_pts,1:vel_pts,1:cl_pts,1:bbv_npars))
      allocate(bbv_spls%sp2(1:od_pts,1:alpha_pts,1:vel_pts,1:cl_pts,1:bbv_npars))
      
      call setup_parsplines(bbv_npars,bbv_spls,bbv_par_file)
      bb_splined(2)=.true.
      if (sf_debug) write(*,*) 'BB vector splines set up'
    case (3)  ! tensors
      allocate(bbt_spls)
      allocate(bbt_spls%sp(1:od_pts,1:alpha_pts,1:vel_pts,1:cl_pts,1:bbt_npars))
      allocate(bbt_spls%sp2(1:od_pts,1:alpha_pts,1:vel_pts,1:cl_pts,1:bbt_npars))
      
      call setup_parsplines(bbt_npars,bbt_spls,bbt_par_file)
      bb_splined(3)=.true.
      if (sf_debug) write(*,*) 'BB tensor splines set up'
    case default
      call sf_fatalerror('Bad SF_setupbb call - argument must be 1,2,3')
  end select

end subroutine
!~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
subroutine SF_setupee(svt)
  ! Call subroutines that set up variables for EE calculations
  use sf_vars
  implicit none
  integer, intent(in) :: svt
  integer :: i,j,k,m
  real(8) :: tester,tester_x
  real(8) :: z(1:cl_pts),z2(1:cl_pts)
  
  select case (svt)
    case (1)  ! scalars
      allocate(ees_spls)
      allocate(ees_spls%sp(1:od_pts,1:alpha_pts,1:vel_pts,1:cl_pts,1:ees_npars))
      allocate(ees_spls%sp2(1:od_pts,1:alpha_pts,1:vel_pts,1:cl_pts,1:ees_npars))
      
      call setup_parsplines(ees_npars,ees_spls,ees_par_file)
      ee_splined(1)=.true.
      if (sf_debug) write(*,*) 'EE scalar splines set up'
    case (2)  ! vectors
      allocate(eev_spls)
      allocate(eev_spls%sp(1:od_pts,1:alpha_pts,1:vel_pts,1:cl_pts,1:eev_npars))
      allocate(eev_spls%sp2(1:od_pts,1:alpha_pts,1:vel_pts,1:cl_pts,1:eev_npars))
      
      call setup_parsplines(eev_npars,eev_spls,eev_par_file)
      ee_splined(2)=.true.
      if (sf_debug) write(*,*) 'EE vector splines set up'
    case (3)  ! tensors
      allocate(eet_spls)
      allocate(eet_spls%sp(1:od_pts,1:alpha_pts,1:vel_pts,1:cl_pts,1:eet_npars))
      allocate(eet_spls%sp2(1:od_pts,1:alpha_pts,1:vel_pts,1:cl_pts,1:eet_npars))
      
      call setup_parsplines(eet_npars,eet_spls,eet_par_file)
      ee_splined(3)=.true.
      if (sf_debug) write(*,*) 'EE tensor splines set up'
    case default
      call sf_fatalerror('Bad SF_setupee call - argument must be 1,2,3')
  end select

end subroutine
!~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
subroutine setup_parsplines(npars,spls,fname)
  ! Set up values for spline parameters, read in from params file
  !  (internal routine)
  use sf_vars
  implicit none
  integer :: npars
  type (ParSplines), pointer :: spls
  character(LEN=30) :: fname
  integer i,j,k,m,p
  real(8) :: tester,tester_x,testerd,testerdd
  real(8) :: z(1:cl_pts),z2(1:cl_pts)

  ! Read in file
  open(unit=70,file=trim(fname),status='unknown')
  do p=1,od_pts
    do i=1,alpha_pts
      do j=1,vel_pts
        do k=1,cl_pts
          read(70,*) spls%ods(p),spls%alphas(i),spls%vels(j), &
                     spls%cls(k),spls%sp(p,i,j,k,1:npars)
        end do
      end do
    end do
  end do
  close(70)
  
  ! For each od, alpha, and vel value, make spline of each parameter
  !  w.r.t. cl values
  do p=1,od_pts
    do i=1,alpha_pts
      do j=1,vel_pts
        do k=1,npars
          z(1:cl_pts)=spls%sp(p,i,j,1:cl_pts,k)
          call spline_cubic_set(cl_pts,spls%cls,z,2,0.0d0,2,0.0d0,z2)
          spls%sp2(p,i,j,1:cl_pts,k)=z2(1:cl_pts)
        end do
      end do
    end do
  end do

end subroutine
!~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
subroutine SF_finishbb()
  ! Deallocate memory allocated by SF_setupbb
  use sf_vars
  implicit none
  
  if (bb_splined(2)) then
    deallocate(bbv_spls%sp,bbv_spls%sp2)
    deallocate(bbv_spls)
    bb_splined(2)=.false.
  end if
  if (bb_splined(3)) then
    deallocate(bbt_spls%sp,bbt_spls%sp2)
    deallocate(bbt_spls)
    bb_splined(3)=.false.
  end if
  
end subroutine
!~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
subroutine SF_finishtt()
  ! Deallocate memory allocated by SF_setupbb
  use sf_vars
  implicit none
  
  if (tt_splined(1)) then
    deallocate(tts_spls%sp,tts_spls%sp2)
    deallocate(tts_spls)
    tt_splined(1)=.false.
  end if
  if (tt_splined(2)) then
    deallocate(ttv_spls%sp,ttv_spls%sp2)
    deallocate(ttv_spls)
    tt_splined(2)=.false.
  end if
  if (tt_splined(3)) then
    deallocate(ttt_spls%sp,ttt_spls%sp2)
    deallocate(ttt_spls)
    tt_splined(3)=.false.
  end if
  
end subroutine
!~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
subroutine SF_finishee()
  ! Deallocate memory allocated by SF_setupee
  use sf_vars
  implicit none
  
  if (ee_splined(1)) then
    deallocate(ees_spls%sp,ees_spls%sp2)
    deallocate(ees_spls)
    ee_splined(1)=.false.
  end if
  if (ee_splined(2)) then
    deallocate(eev_spls%sp,eev_spls%sp2)
    deallocate(eev_spls)
    ee_splined(2)=.false.
  end if
  if (ee_splined(3)) then
    deallocate(eet_spls%sp,eet_spls%sp2)
    deallocate(eet_spls)
    ee_splined(3)=.false.
  end if
  
end subroutine
!~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
real(8) function SF_ttcalc(svt,gmu,od,alpha,vel,cl,ell)
  ! Call subroutines that calculate TT C_l's
  use sf_vars
  implicit none
  real(8) :: SF_ttcalc
  integer, intent(in) :: svt,ell
  real(8), intent(in) :: gmu,od,alpha,vel,cl
  
  if (.not.tt_splined(svt)) then
    call sf_fatalerror('need to call SF_setuptt before SF_ttcalc!')
  end if
  
  call check_input_pars(od,alpha,vel,cl)
  
  select case (svt)
    case (1)  ! scalars
      SF_ttcalc=ttscalc(gmu,od,alpha,vel,cl,ell)
    case (2)  ! vectors
      SF_ttcalc=ttvcalc(gmu,od,alpha,vel,cl,ell)
    case (3)  ! tensors
      SF_ttcalc=tttcalc(gmu,od,alpha,vel,cl,ell)
  end select

end function
!~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
real(8) function SF_bbcalc(svt,gmu,od,alpha,vel,cl,ell)
  ! Call subroutines that calculate BB C_l's
  use sf_vars
  implicit none
  real(8) :: SF_bbcalc
  integer, intent(in) :: svt,ell
  real(8), intent(in) :: gmu,od,alpha,vel,cl
  
  if ((svt/=2).and.(svt/=3)) then
    call sf_fatalerror('Bad SF_bbcalc call - need svt=2,3')
  end if
  if (.not.bb_splined(svt)) then
    call sf_fatalerror('need to call SF_setupbb before SF_bbcalc!')
  end if
  
  call check_input_pars(od,alpha,vel,cl)
  
  select case (svt)
    case (2)  ! vectors
      SF_bbcalc=bbvcalc(gmu,od,alpha,vel,cl,ell)
    case (3)  ! tensors
      SF_bbcalc=bbtcalc(gmu,od,alpha,vel,cl,ell)
  end select

end function
!~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
real(8) function SF_eecalc(svt,gmu,od,alpha,vel,cl,ell)
  ! Call subroutines that calculate EE C_l's
  use sf_vars
  implicit none
  real(8) :: SF_eecalc
  integer, intent(in) :: svt,ell
  real(8), intent(in) :: gmu,od,alpha,vel,cl
  
  if (.not.ee_splined(svt)) then
    call sf_fatalerror('need to call SF_setupee before SF_eecalc!')
  end if
  
  call check_input_pars(od,alpha,vel,cl)
  
  select case (svt)
    case (1)  ! scalars
      SF_eecalc=eescalc(gmu,od,alpha,vel,cl,ell)
    case (2)  ! vectors
      SF_eecalc=eevcalc(gmu,od,alpha,vel,cl,ell)
    case (3)  ! tensors
      SF_eecalc=eetcalc(gmu,od,alpha,vel,cl,ell)
  end select

end function
!~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
real(8) function ttscalc(gmu,od,alpha,vel,cl,ell)
  ! Calculate l(l+1)/twopi*Tcmb^2*C_l^TT (scalar mode) for specified ell
  !  (internal routine)
  use sf_vars
  implicit none
  real(8) :: ttscalc
  integer, intent(in) :: ell
  real(8), intent(in) :: gmu,od,alpha,vel,cl
  logical :: same_pars
  real(8) :: temp_pars(1:tts_npars)
  integer :: i
  real(8) :: ell_real,der1,der2
  
  real(8), save :: od_last=0.0d0
  real(8), save :: alpha_last=0.0d0
  real(8), save :: vel_last=0.0d0
  real(8), save :: cl_last=0.0d0
  
  integer, save :: pt1,pt2,pt3
  real(8), save :: Cl1
  real(8), save :: TTspl(tts_splpts),TTspl2(tts_splpts)
  real(8), save :: TTspl_x_real(1:tts_splpts)
  integer, save :: TTspl_x(1:tts_splpts)
  real(8), save :: TTspl_ld,TTspl_rd
  real(8), save :: p1b,p1c,p1d,p1e,p1f,p2b,p2c,p2d,p2e
  real(8), save :: p3b,p3c,p3d,p3e
  
  ! If string pars are the same as the last call, don't re-evaluate
  !  splines
  same_pars=.true.
  if ((abs(od-od_last)>par_tol).or. &
      (abs(alpha-alpha_last)>par_tol).or. &
      (abs(vel-vel_last)>par_tol).or. &
      (abs(cl-cl_last)>par_tol)) then
    od_last=od
    alpha_last=alpha
    vel_last=vel
    cl_last=cl
    same_pars=.false.
  end if
  
  if (.not.same_pars) then
    call get_temp_pars(od,alpha,vel,cl,tts_spls%ods,tts_spls%alphas, &
                       tts_spls%vels,tts_spls%cls,tts_npars, &
                       tts_spls%sp,tts_spls%sp2,temp_pars)
    
    ! This helps with readability of the fitting functions later on
    pt1=temp_pars(1)
    pt2=temp_pars(2)
    pt3=temp_pars(3)
    p1b=temp_pars(4)
    p1c=temp_pars(5)
    p1d=temp_pars(6)
    p1e=temp_pars(7)
    p1f=temp_pars(8)
    p2b=temp_pars(9)
    p2c=temp_pars(10)
    p2d=temp_pars(11)
    p2e=temp_pars(12)
    p3b=temp_pars(13)
    p3c=temp_pars(14)
    p3d=temp_pars(15)
    p3e=temp_pars(16)
    TTspl_ld=temp_pars(17)
    TTspl_rd=temp_pars(18)
    Cl1=temp_pars(19)
    do i=1,tts_splpts
      TTspl(i)=temp_pars(i+19)
    end do
    
    ! Spline for acoustic peak region
    do i=1,tts_splpts
      TTspl_x(i)=pt2+(pt3-pt2)*(i-1)/(tts_splpts-1)
      TTspl_x_real(i)=TTspl_x(i)
    end do
    call spline_cubic_set(tts_splpts,TTspl_x_real,TTspl, &
                          1,TTspl_ld,1,TTspl_rd,TTspl2)
  end if
  
  if (ell<=pt1) then
    ttscalc=Cl1 &
            +p1b*(ell-pt1) &
            +p1c*(ell-pt1)*(ell-pt1) &
            +p1d*(ell-pt1)*(ell-pt1)*(ell-pt1) &
            +p1e*(ell-pt1)*(ell-pt1)*(ell-pt1)*(ell-pt1) &
            +p1f*(ell-pt1)*log(dble(ell))
  
  else if ((ell>pt1).and.(ell<=pt2)) then
    ttscalc=Cl1 &
            +p2b*(ell-pt1)+p2c*(ell-pt1)*(ell-pt1) &
            +p2d*(ell-pt1)*(ell-pt1)*(ell-pt1) &
            +p2e*(ell-pt1)*(ell-pt1)*(ell-pt1)*(ell-pt1)
  
  else if ((ell>pt2).and.(ell<=pt3)) then
    ell_real=ell
    call spline_cubic_val(tts_splpts,TTspl_x_real,TTspl,TTspl2, &
                          ell_real,ttscalc,der1,der2)
  
  else
    ttscalc=TTspl(tts_splpts) &
            +p3b*(ell-pt3)+p3c*(ell-pt3)*(ell-pt3) &
            +p3d*(ell-pt3)*(ell-pt3)*(ell-pt3) &
            +p3e*(ell-pt3)*(ell-pt3)*(ell-pt3)*(ell-pt3)
  
  end if

  ttscalc=ttscalc*gmu*gmu/gmu_grid/gmu_grid

end function
!~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
real(8) function ttvcalc(gmu,od,alpha,vel,cl,ell)
  ! Calculate l(l+1)/twopi*Tcmb^2*C_l^TT (vector mode) for specified ell
  !  (internal routine)
  use sf_vars
  implicit none
  real(8) :: ttvcalc
  integer, intent(in) :: ell
  real(8), intent(in) :: gmu,od,alpha,vel,cl
  logical :: same_pars
  real(8) :: temp_pars(1:ttv_npars)
  integer :: i
  real(8) :: ell_real,der1,der2
  integer :: aend,bend,cend,dend

  real(8), save :: od_last=0.0d0
  real(8), save :: alpha_last=0.0d0
  real(8), save :: vel_last=0.0d0
  real(8), save :: cl_last=0.0d0

  integer, save :: pt1,pt2,pt3
  real(8), save :: Cl1
  real(8), save :: TTspl(ttv_splpts),TTspl2(ttv_splpts)
  real(8), save :: TTspl_x_real(1:ttv_splpts)
  integer, save :: TTspl_x(1:ttv_splpts)
  real(8), save :: TTspl_ld,TTspl_rd
  
  real(8), save :: p1b,p1c,p1d,p1e,p1f,p2b,p2c,p2d,p2e,p2f
  real(8), save :: p3b,p3c,p3d
  
  same_pars=.true.
  if ((abs(od-od_last)>par_tol).or. &
      (abs(alpha-alpha_last)>par_tol).or. &
      (abs(vel-vel_last)>par_tol).or. &
      (abs(cl-cl_last)>par_tol)) then
    od_last=od
    alpha_last=alpha
    vel_last=vel
    cl_last=cl
    same_pars=.false.
  end if
  
  if (.not.same_pars) then
    call get_temp_pars(od,alpha,vel,cl,ttv_spls%ods,ttv_spls%alphas, &
                       ttv_spls%vels,ttv_spls%cls,ttv_npars, &
                       ttv_spls%sp,ttv_spls%sp2,temp_pars)
    
    TTspl_ld=temp_pars(1)
    TTspl_rd=temp_pars(2)
    do i=1,ttv_splpts
      TTspl(i)=temp_pars(i+2)
    end do
    
    ! Spline over whole region
    TTspl_x(1)=3
    aend=ttv_splpts_a
    bend=aend+ttv_splpts_b
    cend=bend+ttv_splpts_c
    dend=cend+ttv_splpts_d
    
    do i=1,aend
      TTspl_x(i)=3+(200-3)*(i-1)/(aend-1)
      TTspl_x_real(i)=TTspl_x(i)
    end do
    do i=aend,bend
      TTspl_x(i)=200+(1050-200)*(i-aend)/(bend-aend)
      TTspl_x_real(i)=TTspl_x(i)
    end do
    do i=bend,cend
      TTspl_x(i)=1050+(1400-1050)*(i-bend)/(cend-bend)
      TTspl_x_real(i)=TTspl_x(i)
    end do
    do i=cend,dend
      TTspl_x(i)=1400+(2997-1400)*(i-cend)/(dend-cend)
      TTspl_x_real(i)=TTspl_x(i)
    end do
    call spline_cubic_set(ttv_splpts,TTspl_x_real,TTspl, &
                          1,TTspl_ld,1,TTspl_rd,TTspl2)
    
  end if
  
  ell_real=ell
  call spline_cubic_val(ttv_splpts,TTspl_x_real,TTspl,TTspl2, &
                          ell_real,ttvcalc,der1,der2)
  
  ttvcalc=ttvcalc*gmu*gmu/gmu_grid/gmu_grid

end function
!~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
real(8) function tttcalc(gmu,od,alpha,vel,cl,ell)
  ! Calculate l(l+1)/twopi*Tcmb^2*C_l^TT (tensor mode) for specified ell
  !  (internal routine)
  use sf_vars
  implicit none
  real(8) :: tttcalc
  integer, intent(in) :: ell
  real(8), intent(in) :: gmu,od,alpha,vel,cl
  logical :: same_pars
  real(8) :: temp_pars(1:ttt_npars)
  integer :: i,j,k,m
  real(8) :: ell_real,der1,der2
  
  real(8), save :: od_last=0.0d0
  real(8), save :: alpha_last=0.0d0
  real(8), save :: vel_last=0.0d0
  real(8), save :: cl_last=0.0d0
  
  integer, save :: pt1,pt2,pt3,pt4
  real(8), save :: Cl1
  real(8), save :: Cl_2,Cl_3,Cl_4,Cl_5
  real(8), save :: TTspl(ttt_splpts),TTspl2(ttt_splpts)
  real(8), save :: TTspl_x_real(1:ttt_splpts)
  integer, save :: TTspl_x(1:ttt_splpts)
  real(8), save :: TTspl_ld,TTspl_rd
  real(8), save :: p1b,p1c,p1d,p1e,p2b,p2c,p2d,p2e
  real(8), save :: p3b,p3c,p3d,p4b,p4c,p4d
  
  same_pars=.true.
  if ((abs(od-od_last)>par_tol).or. &
      (abs(alpha-alpha_last)>par_tol).or. &
      (abs(vel-vel_last)>par_tol).or. &
      (abs(cl-cl_last)>par_tol)) then
    od_last=od
    alpha_last=alpha
    vel_last=vel
    cl_last=cl
    same_pars=.false.
  end if
  
  if (.not.same_pars) then
    call get_temp_pars(od,alpha,vel,cl,ttt_spls%ods,ttt_spls%alphas, &
                       ttt_spls%vels,ttt_spls%cls,ttt_npars, &
                       ttt_spls%sp,ttt_spls%sp2,temp_pars)
    
    ! This helps with readability of the fitting functions later on
    pt1=temp_pars(1)
    pt2=temp_pars(2)
    pt3=temp_pars(3)
    pt4=temp_pars(4)
    p1b=temp_pars(5)
    p1c=temp_pars(6)
    p1d=temp_pars(7)
    p1e=temp_pars(8)
    p2b=temp_pars(9)
    p2c=temp_pars(10)
    p2d=temp_pars(11)
    p2e=temp_pars(12)
    p3b=temp_pars(13)
    p3c=temp_pars(14)
    p3d=temp_pars(15)
    p4b=temp_pars(16)
    p4c=temp_pars(17)
    p4d=temp_pars(18)
    
    TTspl_ld=temp_pars(19)
    TTspl_rd=temp_pars(20)
    Cl1=temp_pars(21)
    Cl_2=temp_pars(22)
    Cl_3=temp_pars(23)
    Cl_4=temp_pars(24)
    Cl_5=temp_pars(25)
    do i=1,ttt_splpts
      TTspl(i)=temp_pars(i+25)
    end do
    
    ! Spline for possibly nontrivial region
    do i=1,ttt_splpts
      TTspl_x(i)=pt3+(pt4-pt3)*(i-1)/(ttt_splpts-1)
      TTspl_x_real(i)=TTspl_x(i)
    end do
    call spline_cubic_set(ttt_splpts,TTspl_x_real,TTspl, &
                          1,TTspl_ld,1,TTspl_rd,TTspl2)
  end if
  
  if (ell==2) then
    tttcalc=Cl_2
    
  else if (ell==3) then
    tttcalc=Cl_3
    
  else if (ell==4) then
    tttcalc=Cl_4
    
  else if (ell==5) then
    tttcalc=Cl_5
    
  else if ((ell>5).and.(ell<=pt1)) then
    tttcalc=Cl1 &
            +p1b*(ell-pt1) &
            +p1c*(ell-pt1)*(ell-pt1) &
            +p1d*(ell-pt1)*(ell-pt1)*(ell-pt1) &
            +p1e*(ell-pt1)*(ell-pt1)*(ell-pt1)*(ell-pt1)
  
  else if ((ell>pt1).and.(ell<=pt2)) then
    tttcalc=Cl1 &
            +p2b*(ell-pt1) &
            +p2c*(ell-pt1)*(ell-pt1) &
            +p2d*(ell-pt1)*(ell-pt1)*(ell-pt1) &
            +p2e*(ell-pt1)*(ell-pt1)*(ell-pt1)*(ell-pt1)
  
  else if ((ell>pt2).and.(ell<=pt3)) then
    tttcalc=TTspl(1) &
            +p3b*(ell-pt3) &
            +p3c*(ell-pt3)/ell/ell &
            +p3d*(ell-pt3)/ell
  
  else if ((ell>pt3).and.(ell<=pt4)) then
    ell_real=ell
    call spline_cubic_val(ttt_splpts,TTspl_x_real,TTspl,TTspl2, &
                          ell_real,tttcalc,der1,der2)
  else
    tttcalc=TTspl(ttt_splpts) &
            +p4b*(ell-pt4) &
            +p4c*(ell-pt4)/ell/ell &
            +p4d*(ell-pt4)/ell
  
  end if

  tttcalc=tttcalc*gmu*gmu/gmu_grid/gmu_grid

end function
!~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
real(8) function bbvcalc(gmu,od,alpha,vel,cl,ell)
  ! Calculate l(l+1)/twopi*Tcmb^2*C_l^BB (vector model) for specified ell
  !  (interla routine)
  use sf_vars
  implicit none
  real(8) :: bbvcalc
  integer, intent(in) :: ell
  real(8), intent(in) :: gmu,od,alpha,vel,cl
  logical :: same_pars
  real(8) :: temp_pars(1:bbv_npars)
  integer :: i,j,k,m
  real(8) :: ell_real,der1,der2
  
  real(8), save :: od_last=0.0d0
  real(8), save :: alpha_last=0.0d0
  real(8), save :: vel_last=0.0d0
  real(8), save :: cl_last=0.0d0

  real(8), save :: BBspl(bbv_splpts),BBspl2(bbv_splpts)
  real(8), save :: BBspl_x_real(1:bbv_splpts)
  integer, save :: BBspl_x(1:bbv_splpts)
  real(8), save :: BBspl_ld,BBspl_rd

  integer, save :: pt1,pt2,pt3
  real(8), save :: Cl1
  real(8), save :: p1b,p1c,p1d,p1e,p1f,p2b,p2c,p2d,p2e
  real(8), save :: p4b,p4c,p4d,p4e,p5b,p5c,p5d,p5e

  same_pars=.true.
  if ((abs(od-od_last)>par_tol).or. &
      (abs(alpha-alpha_last)>par_tol).or. &
      (abs(vel-vel_last)>par_tol).or. &
      (abs(cl-cl_last)>par_tol)) then
    od_last=od
    alpha_last=alpha
    vel_last=vel
    cl_last=cl
    same_pars=.false.
  end if
  
  if (.not.same_pars) then
    call get_temp_pars(od,alpha,vel,cl,bbv_spls%ods,bbv_spls%alphas, &
                       bbv_spls%vels,bbv_spls%cls,bbv_npars, &
                       bbv_spls%sp,bbv_spls%sp2,temp_pars)
    
    BBspl_ld=temp_pars(1)
    BBspl_rd=temp_pars(2)
    do i=1,bbv_splpts
      BBspl(i)=temp_pars(i+2)
    end do
    
    ! Spline over whole region
    BBspl_x(1)=3
    BBspl_x(2)=5
    BBspl_x(3)=10
    BBspl_x(4)=15
    BBspl_x(5)=25
    BBspl_x(6)=40
    BBspl_x(7)=70
    BBspl_x(8)=120
    BBspl_x_real(1:8)=BBspl_x(1:8)
    do i=8,bbv_splpts
      BBspl_x(i)=120+(2997-120)*(i-8)/(bbv_splpts-8)
      BBspl_x_real(i)=BBspl_x(i)
    end do
    call spline_cubic_set(bbv_splpts,BBspl_x_real,BBspl, &
                          1,BBspl_ld,1,BBspl_rd,BBspl2)
    
  end if
  
  ell_real=ell
  call spline_cubic_val(bbv_splpts,BBspl_x_real,BBspl,BBspl2, &
                          ell_real,bbvcalc,der1,der2)
  
  bbvcalc=bbvcalc*gmu*gmu/gmu_grid/gmu_grid

end function
!~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
real(8) function bbtcalc(gmu,od,alpha,vel,cl,ell)
  ! Calculate l(l+1)/twopi*Tcmb^2*C_l^BB (tensor mode) for specified ell
  !  (internal routine)
  use sf_vars
  implicit none
  real(8) :: bbtcalc
  integer, intent(in) :: ell
  real(8), intent(in) :: gmu,od,alpha,vel,cl
  logical :: same_pars
  real(8) :: temp_pars(1:bbt_npars)
  integer :: i
  real(8) :: ell_real,der1,der2

  real(8), save :: od_last=0.0d0
  real(8), save :: alpha_last=0.0d0
  real(8), save :: vel_last=0.0d0
  real(8), save :: cl_last=0.0d0
  
  real(8), save :: BBspl(bbt_splpts),BBspl2(bbt_splpts)
  real(8), save :: BBspl_x_real(1:bbt_splpts)
  integer, save :: BBspl_x(1:bbt_splpts)
  real(8), save :: BBspl_ld,BBspl_rd

  integer, save :: pt1,pt2,pt3
  real(8), save :: Cl1
  real(8), save :: p1b,p1c,p1d,p1e,p1f,p2b,p2c,p2d,p2e
  real(8), save :: p3b,p3c,p3d,p3e,p3f,p4b,p4c,p4d

  same_pars=.true.
  if ((abs(od-od_last)>par_tol).or. &
      (abs(alpha-alpha_last)>par_tol).or. &
      (abs(vel-vel_last)>par_tol).or. &
      (abs(cl-cl_last)>par_tol)) then
    od_last=od
    alpha_last=alpha
    vel_last=vel
    cl_last=cl
    same_pars=.false.
  end if
  
  if (.not.same_pars) then
    call get_temp_pars(od,alpha,vel,cl,bbt_spls%ods,bbt_spls%alphas, &
                       bbt_spls%vels,bbt_spls%cls,bbt_npars, &
                       bbt_spls%sp,bbt_spls%sp2,temp_pars)
        
    BBspl_ld=temp_pars(1)
    BBspl_rd=temp_pars(2)
    do i=1,bbt_splpts
      BBspl(i)=temp_pars(i+2)
    end do
    
    ! Spline over whole region
    BBspl_x(1)=3
    BBspl_x(2)=5
    BBspl_x(3)=8
    BBspl_x(4)=10
    BBspl_x(5)=15
    BBspl_x(6)=20
    BBspl_x(7)=30
    BBspl_x_real(1:7)=BBspl_x(1:7)
    
    do i=7,7+bbt_splpts_a
      BBspl_x(i)=30+(800-30)*(i-7)/(bbt_splpts_a)
      BBspl_x_real(i)=BBspl_x(i)
    end do
    do i=7+bbt_splpts_a,7+bbt_splpts_a+bbt_splpts_b
      BBspl_x(i)=800+(2997-800)*(i-(7+bbt_splpts_a))/(bbt_splpts_b)
      BBspl_x_real(i)=BBspl_x(i)
    end do
    call spline_cubic_set(bbt_splpts,BBspl_x_real,BBspl, &
                          1,BBspl_ld,1,BBspl_rd,BBspl2)
  end if
  
  ell_real=ell
  call spline_cubic_val(bbt_splpts,BBspl_x_real,BBspl,BBspl2, &
                        ell_real,bbtcalc,der1,der2)

  bbtcalc=bbtcalc*gmu*gmu/gmu_grid/gmu_grid

end function
!~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
real(8) function eescalc(gmu,od,alpha,vel,cl,ell)
  ! Calculate l(l+1)/twopi*Tcmb^2*C_l^EE (scalar mode) for specified ell
  !  (internal routine)
  use sf_vars
  implicit none
  real(8) :: eescalc
  integer, intent(in) :: ell
  real(8), intent(in) :: gmu,od,alpha,vel,cl
  logical :: same_pars
  real(8) :: temp_pars(1:ees_npars)
  integer :: i
  real(8) :: ell_real,der1,der2
  
  real(8), save :: od_last=0.0d0
  real(8), save :: alpha_last=0.0d0
  real(8), save :: vel_last=0.0d0
  real(8), save :: cl_last=0.0d0
  
  integer, save :: pt1,pt2,pt3
  real(8), save :: Cl1
  real(8), save :: EEspl(ees_splpts),EEspl2(ees_splpts)
  real(8), save :: EEspl_x_real(1:ees_splpts)
  integer, save :: EEspl_x(1:ees_splpts)
  real(8), save :: EEspl_ld,EEspl_rd
  real(8), save :: p1b,p1c,p1d,p1e,p1f,p2b,p2c,p2d,p2e
  real(8), save :: p3b,p3c,p3d,p3e
  
  same_pars=.true.
  if ((abs(od-od_last)>par_tol).or. &
      (abs(alpha-alpha_last)>par_tol).or. &
      (abs(vel-vel_last)>par_tol).or. &
      (abs(cl-cl_last)>par_tol)) then
    od_last=od
    alpha_last=alpha
    vel_last=vel
    cl_last=cl
    same_pars=.false.
  end if
  
  if (.not.same_pars) then
    call get_temp_pars(od,alpha,vel,cl,ees_spls%ods,ees_spls%alphas, &
                       ees_spls%vels,ees_spls%cls,ees_npars, &
                       ees_spls%sp,ees_spls%sp2,temp_pars)
    
    EEspl_ld=temp_pars(1)
    EEspl_rd=temp_pars(2)
    do i=1,ees_splpts
      EEspl(i)=temp_pars(i+2)
    end do
    
    ! Spline over whole region
    EEspl_x(1)=3
    EEspl_x(2)=10
    EEspl_x(3)=20
    EEspl_x(4)=35
    EEspl_x(5)=50
    EEspl_x(6)=75
    EEspl_x_real(1:6)=EEspl_x(1:6)
    do i=6,ees_splpts
      EEspl_x(i)=75+(2997-75)*(i-6)/(ees_splpts-6)
      EEspl_x_real(i)=EEspl_x(i)
    end do
    call spline_cubic_set(ees_splpts,EEspl_x_real,EEspl, &
                          1,EEspl_ld,1,EEspl_rd,EEspl2)
  end if
  
  ell_real=ell
  call spline_cubic_val(ees_splpts,EEspl_x_real,EEspl,EEspl2, &
                          ell_real,eescalc,der1,der2)

  eescalc=eescalc*gmu*gmu/gmu_grid/gmu_grid

end function
!~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
real(8) function eevcalc(gmu,od,alpha,vel,cl,ell)
  ! Calculate l(l+1)/twopi*Tcmb^2*C_l^EE (vector mode) for specified ell
  !  (internal routine)
  use sf_vars
  implicit none
  real(8) :: eevcalc
  integer, intent(in) :: ell
  real(8), intent(in) :: gmu,od,alpha,vel,cl
  logical :: same_pars
  real(8) :: temp_pars(1:eev_npars)
  integer :: i
  real(8) :: ell_real,der1,der2

  real(8), save :: od_last=0.0d0
  real(8), save :: alpha_last=0.0d0
  real(8), save :: vel_last=0.0d0
  real(8), save :: cl_last=0.0d0

  real(8), save :: EEspl(eev_splpts),EEspl2(eev_splpts)
  real(8), save :: EEspl_x_real(1:eev_splpts)
  integer, save :: EEspl_x(1:eev_splpts)
  real(8), save :: EEspl_ld,EEspl_rd

  integer, save :: pt1,pt2,pt3
  real(8), save :: Cl1
  real(8), save :: p1b,p1c,p1d,p1e,p1f,p2b,p2c,p2d,p2e
  real(8), save :: p3b,p3c,p3d,p3e,p4b,p4c,p4d

  same_pars=.true.
  if ((abs(od-od_last)>par_tol).or. &
      (abs(alpha-alpha_last)>par_tol).or. &
      (abs(vel-vel_last)>par_tol).or. &
      (abs(cl-cl_last)>par_tol)) then
    od_last=od
    alpha_last=alpha
    vel_last=vel
    cl_last=cl
    same_pars=.false.
  end if

  if (.not.same_pars) then
    call get_temp_pars(od,alpha,vel,cl,eev_spls%ods,eev_spls%alphas, &
                       eev_spls%vels,eev_spls%cls,eev_npars, &
                       eev_spls%sp,eev_spls%sp2,temp_pars)

    EEspl_ld=temp_pars(1)
    EEspl_rd=temp_pars(2)
    do i=1,eev_splpts
      EEspl(i)=temp_pars(i+2)
    end do
    
    ! Spline over whole region
    EEspl_x(1)=3
    EEspl_x(2)=5
    EEspl_x(3)=10
    EEspl_x(4)=15
    EEspl_x(5)=20
    EEspl_x(6)=30
    EEspl_x(7)=40
    EEspl_x(8)=50
    EEspl_x(9)=100
    EEspl_x_real(1:9)=EEspl_x(1:9)
    do i=9,eev_splpts
      EEspl_x(i)=100+(2997-100)*(i-9)/(eev_splpts-9)
      EEspl_x_real(i)=EEspl_x(i)
    end do
    call spline_cubic_set(eev_splpts,EEspl_x_real,EEspl, &
                          1,EEspl_ld,1,EEspl_rd,EEspl2)
  end if
  
  ell_real=ell
  call spline_cubic_val(eev_splpts,EEspl_x_real,EEspl,EEspl2, &
                          ell_real,eevcalc,der1,der2)
  
  eevcalc=eevcalc*gmu*gmu/gmu_grid/gmu_grid

end function
!~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
real(8) function eetcalc(gmu,od,alpha,vel,cl,ell)
  ! Calculate l(l+1)/twopi*Tcmb^2*C_l^EE (tensor mode) for specified ell
  !  (internal routine)
  use sf_vars
  implicit none
  real(8) :: eevcalc
  integer, intent(in) :: ell
  real(8), intent(in) :: gmu,od,alpha,vel,cl
  logical :: same_pars
  real(8) :: temp_pars(1:eet_npars)
  integer :: i
  real(8) :: ell_real,der1,der2

  real(8), save :: od_last=0.0d0
  real(8), save :: alpha_last=0.0d0
  real(8), save :: vel_last=0.0d0
  real(8), save :: cl_last=0.0d0

  real(8), save :: EEspl(eet_splpts),EEspl2(eet_splpts)
  real(8), save :: EEspl_x_real(1:eet_splpts)
  integer, save :: EEspl_x(1:eet_splpts)
  real(8), save :: EEspl_ld,EEspl_rd

  integer, save :: pt1,pt2,pt3
  real(8), save :: Cl1
  real(8), save :: p1b,p1c,p1d,p1e,p1f,p2b,p2c,p2d,p2e
  real(8), save :: p3b,p3c,p3d,p3e,p3f,p4b,p4c,p4d

  same_pars=.true.
  if ((abs(od-od_last)>par_tol).or. &
      (abs(alpha-alpha_last)>par_tol).or. &
      (abs(vel-vel_last)>par_tol).or. &
      (abs(cl-cl_last)>par_tol)) then
    od_last=od
    alpha_last=alpha
    vel_last=vel
    cl_last=cl
    same_pars=.false.
  end if

  if (.not.same_pars) then
    call get_temp_pars(od,alpha,vel,cl,eet_spls%ods,eet_spls%alphas, &
                       eet_spls%vels,eet_spls%cls,eet_npars, &
                       eet_spls%sp,eet_spls%sp2,temp_pars)
    
    EEspl_ld=temp_pars(1)
    EEspl_rd=temp_pars(2)
    do i=1,eet_splpts
      EEspl(i)=temp_pars(i+2)
    end do
    
    ! Spline over whole region
    EEspl_x(1)=3
    EEspl_x(2)=5
    EEspl_x(3)=7
    EEspl_x(4)=10
    EEspl_x(5)=13
    EEspl_x(6)=17
    EEspl_x(7)=20
    EEspl_x(8)=30
    EEspl_x(9)=70
    EEspl_x_real(1:9)=EEspl_x(1:9)
    
    do i=9,9+eet_splpts_a
      EEspl_x(i)=70+(1000-70)*(i-9)/(eet_splpts_a)
      EEspl_x_real(i)=EEspl_x(i)
    end do
    do i=9+eet_splpts_a,9+eet_splpts_a+eet_splpts_b
      EEspl_x(i)=1000+(2997-1000)*(i-(9+eet_splpts_a))/(eet_splpts_b)
      EEspl_x_real(i)=EEspl_x(i)
    end do
    call spline_cubic_set(eet_splpts,EEspl_x_real,EEspl, &
                          1,EEspl_ld,1,EEspl_rd,EEspl2)
  end if
  
  ell_real=ell
  call spline_cubic_val(eet_splpts,EEspl_x_real,EEspl,EEspl2, &
                          ell_real,eetcalc,der1,der2)
  
  eetcalc=eetcalc*gmu*gmu/gmu_grid/gmu_grid

end function
!~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
subroutine get_temp_pars(od,alpha,vel,cl,od_arr,alpha_arr,vel_arr, &
                         cl_arr,np,sp,sp2,temp_pars)
  ! Get fit parameters as a function of od,alpha,vel,cl using 4-dim.
  !  splines (internal routine)
  use sf_vars
  implicit none
  real(8), intent(in) :: od,alpha,vel,cl
  real(8), intent(in) :: od_arr(1:od_pts),alpha_arr(1:alpha_pts), &
                         vel_arr(1:vel_pts),cl_arr(1:cl_pts)
  real(8), intent(in) :: sp(1:od_pts,1:alpha_pts,1:vel_pts,1:cl_pts,1:np)
  real(8), intent(in) :: sp2(1:od_pts,1:alpha_pts,1:vel_pts,1:cl_pts,1:np)
  integer, intent(in) :: np
  real(8), intent(out) :: temp_pars(1:np)
  integer :: i,j,m,p
  real(8) :: y_cl(1:cl_pts),y2_cl(1:cl_pts)
  real(8) :: y_vel(1:vel_pts),y2_vel(1:vel_pts)
  real(8) :: pars_oav(1:od_pts,1:alpha_pts,1:vel_pts,1:np)
  real(8) :: pars_oa(1:od_pts,1:alpha_pts,1:np)
  real(8) :: pars_o(1:od_pts,1:np)
  real(8) :: y_alpha(1:alpha_pts),y2_alpha(1:alpha_pts)
  real(8) :: y_od(1:od_pts),y2_od(1:od_pts)
  real(8) :: der1,der2
  
  ! For each od, alpha and vel, calculate parameter value at specified cl,
  !  and store in pars_oav
  do p=1,od_pts
    do i=1,alpha_pts
      do j=1,vel_pts
        do m=1,np
          y_cl(1:cl_pts)=sp(p,i,j,1:cl_pts,m)
          y2_cl(1:cl_pts)=sp2(p,i,j,1:cl_pts,m)
          call spline_cubic_val(cl_pts,cl_arr,y_cl,y2_cl, &
                                cl,pars_oav(p,i,j,m),der1,der2)
        end do
      end do
    end do
  end do
  
  ! For each od and alpha, calculate parameter value at specified vel,
  !  using parameters in pars_oav to make splines w.r.t. vel
  do p=1,od_pts
    do i=1,alpha_pts
      do m=1,np
        y_vel(1:vel_pts)=pars_oav(p,i,1:vel_pts,m)
        call spline_cubic_set(vel_pts,vel_arr,y_vel, &
                              2,0.0d0,2,0.0d0,y2_vel)
        call spline_cubic_val(vel_pts,vel_arr,y_vel,y2_vel, &
                              vel,pars_oa(p,i,m),der1,der2)      
      end do
    end do
  end do
  
  ! For each od, calculate parameter value at specified alpha,
  !  using parameters in pars_oa to make splines w.r.t. alpha
  do p=1,od_pts
    do m=1,np
      y_alpha(1:alpha_pts)=pars_oa(p,1:alpha_pts,m)
      call spline_cubic_set(alpha_pts,alpha_arr,y_alpha, &
                            2,0.0d0,2,0.0d0,y2_alpha)
      call spline_cubic_val(alpha_pts,alpha_arr,y_alpha,y2_alpha, &
                            alpha,pars_o(p,m),der1,der2)
    end do
  end do
  
  ! Finally, use pars_o to make a spline w.r.t. od, and use
  !  to obtain parameter values at the specified od
  do m=1,np
    y_od(1:od_pts)=pars_o(1:od_pts,m)
    call spline_cubic_set(od_pts,od_arr,y_od, &
                          2,0.0d0,2,0.0d0,y2_od)
    call spline_cubic_val(od_pts,od_arr,y_od,y2_od, &
                          od,temp_pars(m),der1,der2)
  end do

end subroutine
!~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
subroutine check_input_pars(od,alpha,vel,cl)
  ! Check to make sure that input parameters are within grid
  !  (internal routine)
  use sf_vars
  implicit none
  real(8) :: od,alpha,vel,cl
  
  if (((od-grid_od_max)>par_tol).or. &
      ((grid_od_min-od)>par_tol)) then
    call sf_fatalerror('od out of bounds!')
  end if
  if (((alpha-grid_alpha_max)>par_tol).or. &
      ((grid_alpha_min-alpha)>par_tol)) then
    call sf_fatalerror('alpha out of bounds!')
  end if
  if (((vel-grid_vel_max)>par_tol).or. &
      ((grid_vel_min-vel)>par_tol)) then
    call sf_fatalerror('vel out of bounds!')
  end if
  if (((cl-grid_cl_max)>par_tol).or. &
      ((grid_cl_min-cl)>par_tol)) then
    call sf_fatalerror('cl out of bounds!')
  end if

end subroutine
!~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
subroutine sf_fatalerror(str)
  ! Halts program upon fatal error
  !  (internal routine)
  implicit none
	character(LEN=*), intent(in) :: str
  
	write (*,*) 'StringFast error: '//str
	stop 'program terminated - please change code to fix error'
  
end subroutine
!~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
!~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  ! These are public spline routines (and dependencies) written by
  !  John Burkardt, and available at 
  !  http://orion.math.iastate.edu:80/burkardt/f_rc/spline/spline.html
  ! (modified to use double precision)
  
subroutine spline_cubic_set ( n, t, y, ibcbeg, ybcbeg, ibcend, ybcend, ypp )
!
!*******************************************************************************
!
!! SPLINE_CUBIC_SET computes the second derivatives of a cubic spline.
!
!
!  Discussion:
!
!    For data interpolation, the user must call SPLINE_CUBIC_SET to 
!    determine the second derivative data, passing in the data to be 
!    interpolated, and the desired boundary conditions.
!
!    The data to be interpolated, plus the SPLINE_CUBIC_SET output, 
!    defines the spline.  The user may then call SPLINE_CUBIC_VAL to 
!    evaluate the spline at any point.
!
!    The cubic spline is a piecewise cubic polynomial.  The intervals
!    are determined by the "knots" or abscissas of the data to be
!    interpolated.  The cubic spline has continous first and second
!    derivatives over the entire interval of interpolation.  
!
!    For any point T in the interval T(IVAL), T(IVAL+1), the form of
!    the spline is
!
!      SPL(T) = A(IVAL)
!             + B(IVAL) * ( T - T(IVAL) ) 
!             + C(IVAL) * ( T - T(IVAL) )**2
!             + D(IVAL) * ( T - T(IVAL) )**3
!
!    If we assume that we know the values Y(*) and YPP(*), which represent
!    the values and second derivatives of the spline at each knot, then
!    the coefficients can be computed as:
!
!      A(IVAL) = Y(IVAL)
!      B(IVAL) = ( Y(IVAL+1) - Y(IVAL) ) / ( T(IVAL+1) - T(IVAL) )
!        - ( YPP(IVAL+1) + 2 * YPP(IVAL) ) * ( T(IVAL+1) - T(IVAL) ) / 6
!      C(IVAL) = YPP(IVAL) / 2
!      D(IVAL) = ( YPP(IVAL+1) - YPP(IVAL) ) / ( 6 * ( T(IVAL+1) - T(IVAL) ) )
!
!    Since the first derivative of the spline is
!
!      SPL'(T) =     B(IVAL)
!              + 2 * C(IVAL) * ( T - T(IVAL) )
!              + 3 * D(IVAL) * ( T - T(IVAL) )**2,
!
!    the requirement that the first derivative be continuous at interior
!    knot I results in a total of N-2 equations, of the form:
!
!      B(IVAL-1) + 2 C(IVAL-1) * (T(IVAL)-T(IVAL-1)) 
!      + 3 * D(IVAL-1) * (T(IVAL) - T(IVAL-1))**2 = B(IVAL)
!
!    or, setting H(IVAL) = T(IVAL+1) - T(IVAL)
!
!      ( Y(IVAL) - Y(IVAL-1) ) / H(IVAL-1)
!      - ( YPP(IVAL) + 2 * YPP(IVAL-1) ) * H(IVAL-1) / 6
!      + YPP(IVAL-1) * H(IVAL-1)
!      + ( YPP(IVAL) - YPP(IVAL-1) ) * H(IVAL-1) / 2
!      = 
!      ( Y(IVAL+1) - Y(IVAL) ) / H(IVAL)
!      - ( YPP(IVAL+1) + 2 * YPP(IVAL) ) * H(IVAL) / 6
!
!    or
!
!      YPP(IVAL-1) * H(IVAL-1) + 2 * YPP(IVAL) * ( H(IVAL-1) + H(IVAL) )
!      + YPP(IVAL) * H(IVAL) 
!      =
!      6 * ( Y(IVAL+1) - Y(IVAL) ) / H(IVAL)
!      - 6 * ( Y(IVAL) - Y(IVAL-1) ) / H(IVAL-1)    
!
!    Boundary conditions must be applied at the first and last knots.  
!    The resulting tridiagonal system can be solved for the YPP values.
!
!  Modified:
!
!    20 November 2000
!
!  Author:
!
!    John Burkardt
!
!  Parameters:
!
!    Input, integer N, the number of data points; N must be at least 2. 
!
!    Input, real T(N), the points where data is specified.  
!    The values should be distinct, and increasing.
!
!    Input, real Y(N), the data values to be interpolated.
!
!    Input, integer IBCBEG, the left boundary condition flag:
!
!      0: the spline should be a quadratic over the first interval;
!      1: the first derivative at the left endpoint should be YBCBEG;
!      2: the second derivative at the left endpoint should be YBCBEG.
!
!    Input, real YBCBEG, the left boundary value, if needed.
!
!    Input, integer IBCEND, the right boundary condition flag:
!
!      0: the spline should be a quadratic over the last interval;
!      1: the first derivative at the right endpoint should be YBCEND;
!      2: the second derivative at the right endpoint should be YBCEND.
!
!    Input, real YBCEND, the right boundary value, if needed.
!
!    Output, real YPP(N), the second derivatives of the cubic spline.
!
  implicit none
!
  integer n
!
  real(8) diag(n)
  integer i
  integer ibcbeg
  integer ibcend
  real(8) sub(2:n)
  real(8) sup(1:n-1)
  real(8) t(n)
  real(8) y(n)
  real(8) ybcbeg
  real(8) ybcend
  real(8) ypp(n)
!
!  Check.
!
  if ( n <= 1 ) then
    write ( *, '(a)' ) ' '
    write ( *, '(a)' ) 'SPLINE_CUBIC_SET - Fatal error!'
    write ( *, '(a)' ) '  The number of knots must be at least 2.'
    write ( *, '(a,i6)' ) '  The input value of N = ', n
    stop
  end if

  do i = 1, n-1
    if ( t(i) >= t(i+1) ) then
      write ( *, '(a)' ) ' '
      write ( *, '(a)' ) 'SPLINE_CUBIC_SET - Fatal error!'
      write ( *, '(a)' ) '  The knots must be strictly increasing, but'
      write ( *, '(a,i6,a,g14.6)' ) '  T(',  i,') = ', t(i)
      write ( *, '(a,i6,a,g14.6)' ) '  T(',i+1,') = ', t(i+1)
      stop
    end if
  end do
!
!  Set the first equation.
!
  if ( ibcbeg == 0 ) then
    ypp(1) = 0.0d0
    diag(1) = 1.0d0
    sup(1) = -1.0d0
  else if ( ibcbeg == 1 ) then
    ypp(1) = ( y(2) - y(1) ) / ( t(2) - t(1) ) - ybcbeg
    diag(1) = ( t(2) - t(1) ) / 3.0d0 
    sup(1) = ( t(2) - t(1) ) / 6.0d0
  else if ( ibcbeg == 2 ) then
    ypp(1) = ybcbeg
    diag(1) = 1.0d0
    sup(1) = 0.0d0
  else
    write ( *, '(a)' ) ' '
    write ( *, '(a)' ) 'SPLINE_CUBIC_SET - Fatal error!'
    write ( *, '(a)' ) '  The boundary flag IBCBEG must be 0, 1 or 2.'
    write ( *, '(a,i6)' ) '  The input value is IBCBEG = ', ibcbeg
    stop
  end if
!
!  Set the intermediate equations.
!
  do i = 2, n-1
    ypp(i) = ( y(i+1) - y(i) ) / ( t(i+1) - t(i) ) &
           - ( y(i) - y(i-1) ) / ( t(i) - t(i-1) )
    sub(i) = ( t(i) - t(i-1) ) / 6.0d0
    diag(i) = ( t(i+1) - t(i-1) ) / 3.0d0
    sup(i) = ( t(i+1) - t(i) ) / 6.0d0
  end do
!
!  Set the last equation.
!
  if ( ibcend == 0 ) then
    ypp(n) = 0.0d0
    sub(n) = -1.0d0
    diag(n) = 1.0d0
  else if ( ibcend == 1 ) then
    ypp(n) = ybcend - ( y(n) - y(n-1) ) / ( t(n) - t(n-1) )
    sub(n) = ( t(n) - t(n-1) ) / 6.0d0
    diag(n) = ( t(n) - t(n-1) ) / 3.0d0
  else if ( ibcend == 2 ) then
    ypp(n) = ybcend
    sub(n) = 0.0d0
    diag(n) = 1.0d0
  else
    write ( *, '(a)' ) ' '
    write ( *, '(a)' ) 'SPLINE_CUBIC_SET - Fatal error!'
    write ( *, '(a)' ) '  The boundary flag IBCEND must be 0, 1 or 2.'
    write ( *, '(a,i6)' ) '  The input value is IBCEND = ', ibcend
    stop
  end if
!
!  Special case:
!    N = 2, IBCBEG = IBCEND = 0.
!
  if ( n == 2 .and. ibcbeg == 0 .and. ibcend == 0 ) then

    ypp(1) = 0.0d0
    ypp(2) = 0.0d0
!
!  Solve the linear system.
!
  else

    call s3_fs ( sub, diag, sup, n, ypp, ypp )

  end if

  return
end subroutine

subroutine spline_cubic_val ( n, t, y, ypp, tval, yval, ypval, yppval )
!
!*******************************************************************************
!
!! SPLINE_CUBIC_VAL evaluates a cubic spline at a specific point.
!
!
!  Discussion:
!
!    SPLINE_CUBIC_SET must have already been called to define the 
!    values of YPP.
!
!    For any point T in the interval T(IVAL), T(IVAL+1), the form of
!    the spline is
!
!      SPL(T) = A 
!             + B * ( T - T(IVAL) ) 
!             + C * ( T - T(IVAL) )**2
!             + D * ( T - T(IVAL) )**3
!
!    Here:
!      A = Y(IVAL)
!      B = ( Y(IVAL+1) - Y(IVAL) ) / ( T(IVAL+1) - T(IVAL) )
!        - ( YPP(IVAL+1) + 2 * YPP(IVAL) ) * ( T(IVAL+1) - T(IVAL) ) / 6
!      C = YPP(IVAL) / 2
!      D = ( YPP(IVAL+1) - YPP(IVAL) ) / ( 6 * ( T(IVAL+1) - T(IVAL) ) )
!
!  Modified:
!
!    20 November 2000
!
!  Author:
!
!    John Burkardt
!
!  Parameters:
!
!    Input, integer N, the number of data values.
!
!    Input, real T(N), the knot values.
!
!    Input, real Y(N), the data values at the knots.
!
!    Input, real YPP(N), the second derivatives of the spline at the knots.
!
!    Input, real TVAL, a point, typically between T(1) and T(N), at 
!    which the spline is to be evalulated.  If TVAL lies outside 
!    this range, extrapolation is used.
!
!    Output, real YVAL, YPVAL, YPPVAL, the value of the spline, and
!    its first two derivatives at TVAL.
!
  implicit none
!
  integer n
!
  real(8) dt
  real(8) h
  integer left
  integer right
  real(8) t(n)
  real(8) tval
  real(8) y(n)
  real(8) ypp(n)
  real(8) yppval
  real(8) ypval
  real(8) yval
!
!  Determine the interval [T(LEFT), T(RIGHT)] that contains TVAL.
!  Values below T(1) or above T(N) use extrapolation.
!
  call rvec_bracket ( n, t, tval, left, right )
!
!  Evaluate the polynomial.
!
  dt = tval - t(left)
  h = t(right) - t(left)

  yval = y(left) &
       + dt * ( ( y(right) - y(left) ) / h &
              - ( ypp(right) / 6.0d0 + ypp(left) / 3.0d0 ) * h &
       + dt * ( 0.5d0 * ypp(left) &
       + dt * ( ( ypp(right) - ypp(left) ) / ( 6.0d0 * h ) ) ) )

  ypval = ( y(right) - y(left) ) / h &
       - ( ypp(right) / 6.0d0 + ypp(left) / 3.0d0 ) * h &
       + dt * ( ypp(left) &
       + dt * ( 0.5d0 * ( ypp(right) - ypp(left) ) / h ) )

  yppval = ypp(left) + dt * ( ypp(right) - ypp(left) ) / h 

  return
end subroutine

subroutine spline_cubic_val2 ( n, t, y, ypp, left, tval, yval, ypval, yppval )
!
!*******************************************************************************
!
!! SPLINE_CUBIC_VAL2 evaluates a cubic spline at a specific point.
!
!
!  Discussion:
!
!    This routine is a modification of SPLINE_CUBIC_VAL; it allows the
!    user to speed up the code by suggesting the appropriate T interval
!    to search first.
!
!    SPLINE_CUBIC_SET must have already been called to define the
!    values of YPP.
!
!    In the LEFT interval, let RIGHT = LEFT+1.  The form of the spline is
!
!      SPL(T) = 
!          A
!        + B * ( T - T(LEFT) )
!        + C * ( T - T(LEFT) )**2
!        + D * ( T - T(LEFT) )**3
!
!    Here:
!      A = Y(LEFT)
!      B = ( Y(RIGHT) - Y(LEFT) ) / ( T(RIGHT) - T(LEFT) )
!        - ( YPP(RIGHT) + 2 * YPP(LEFT) ) * ( T(RIGHT) - T(LEFT) ) / 6
!      C = YPP(LEFT) / 2
!      D = ( YPP(RIGHT) - YPP(LEFT) ) / ( 6 * ( T(RIGHT) - T(LEFT) ) )
!
!  Modified:
!
!    20 November 2000
!
!  Author:
!
!    John Burkardt
!
!  Parameters:
!
!    Input, integer N, the number of knots.
!
!    Input, real T(N), the knot values.
!
!    Input, real Y(N), the data values at the knots.
!
!    Input, real YPP(N), the second derivatives of the spline at
!    the knots.
!
!    Input/output, integer LEFT, the suggested T interval to search.
!    LEFT should be between 1 and N-1.  If LEFT is not in this range,
!    then its value will be ignored.  On output, LEFT is set to the
!    actual interval in which TVAL lies.
!
!    Input, real TVAL, a point, typically between T(1) and T(N), at
!    which the spline is to be evalulated.  If TVAL lies outside
!    this range, extrapolation is used.
!
!    Output, real YVAL, YPVAL, YPPVAL, the value of the spline, and
!    its first two derivatives at TVAL.
!
  implicit none
!
  integer n
!
  real(8) dt
  real(8) h
  integer left
  integer right
  real(8) t(n)
  real(8) tval
  real(8) y(n)
  real(8) ypp(n)
  real(8) yppval
  real(8) ypval
  real(8) yval
!
!  Determine the interval [T(LEFT), T(RIGHT)] that contains TVAL.
!  
!  What you want from RVEC_BRACKET3 is that TVAL is to be computed
!  by the data in interval {T(LEFT), T(RIGHT)].
!
  left = 0
  call rvec_bracket3 ( n, t, tval, left )
  right = left + 1
!
!  In the interval LEFT, the polynomial is in terms of a normalized
!  coordinate  ( DT / H ) between 0 and 1.
!
  dt = tval - t(left)
  h = t(right) - t(left)

  yval = y(left) + dt * ( ( y(right) - y(left) ) / h &
              - ( ypp(right) / 6.0d0 + ypp(left) / 3.0d0 ) * h &
       + dt * ( 0.5d0 * ypp(left) &
       + dt * ( ( ypp(right) - ypp(left) ) / ( 6.0d0 * h ) ) ) )

  ypval = ( y(right) - y(left) ) / h &
      - ( ypp(right) / 6.0d0 + ypp(left) / 3.0d0 ) * h &
      + dt * ( ypp(left) &
      + dt * ( 0.5d0 * ( ypp(right) - ypp(left) ) / h ) )

  yppval = ypp(left) + dt * ( ypp(right) - ypp(left) ) / h

  return
end subroutine

subroutine rvec_bracket ( n, x, xval, left, right )
!
!*******************************************************************************
!
!! RVEC_BRACKET searches a sorted array for successive brackets of a value.
!
!
!  Discussion:
!
!    If the values in the vector are thought of as defining intervals
!    on the real line, then this routine searches for the interval
!    nearest to or containing the given value.
!
!  Modified:
!
!    06 April 1999
!
!  Author:
!
!    John Burkardt
!
!  Parameters:
!
!    Input, integer N, length of input array.
!
!    Input, real X(N), an array sorted into ascending order.
!
!    Input, real XVAL, a value to be bracketed.
!
!    Output, integer LEFT, RIGHT, the results of the search.
!    Either:
!      XVAL < X(1), when LEFT = 1, RIGHT = 2;
!      XVAL > X(N), when LEFT = N-1, RIGHT = N;
!    or
!      X(LEFT) <= XVAL <= X(RIGHT).
!
  implicit none
!
  integer n
!
  integer i
  integer left
  integer right
  real(8) x(n)
  real(8) xval
!
  do i = 2, n - 1

    if ( xval < x(i) ) then
      left = i - 1
      right = i
      return
    end if

   end do

  left = n - 1
  right = n

  return
end subroutine

subroutine s3_fs ( a1, a2, a3, n, b, x )
!
!*******************************************************************************
!
!! S3_FS factors and solves a tridiagonal linear system.
!
!
!  Note:
!
!    This algorithm requires that each diagonal entry be nonzero.
!
!  Modified:
!
!    05 December 1998
!
!  Author:
!
!    John Burkardt
!
!  Parameters:
!
!    Input/output, real A1(2:N), A2(1:N), A3(1:N-1).
!    On input, the nonzero diagonals of the linear system.
!    On output, the data in these vectors has been overwritten
!    by factorization information.
!
!    Input, integer N, the order of the linear system.
!
!    Input/output, real B(N).
!    On input, B contains the right hand side of the linear system.
!    On output, B has been overwritten by factorization information.
!
!    Output, real X(N), the solution of the linear system.
!
  implicit none
!
  integer n
!
  real(8) a1(2:n)
  real(8) a2(1:n)
  real(8) a3(1:n-1)
  real(8) b(n)
  integer i
  real(8) x(n)
  real(8) xmult
!
!  The diagonal entries can't be zero.
!
  do i = 1, n
    if ( a2(i) == 0.0d0 ) then
      write ( *, '(a)' ) ' '
      write ( *, '(a)' ) 'S3_FS - Fatal error!'
      write ( *, '(a,i6,a)' ) '  A2(', i, ') = 0.'
      return
    end if
  end do

  do i = 2, n-1

    xmult = a1(i) / a2(i-1)
    a2(i) = a2(i) - xmult * a3(i-1)

    b(i) = b(i) - xmult * b(i-1)

  end do

  xmult = a1(n) / a2(n-1)
  a2(n) = a2(n) - xmult * a3(n-1)

  x(n) = ( b(n) - xmult * b(n-1) ) / a2(n)
  do i = n-1, 1, -1
    x(i) = ( b(i) - a3(i) * x(i+1) ) / a2(i)
  end do

  return
end subroutine 

subroutine rvec_bracket3 ( n, t, tval, left )
!
!*******************************************************************************
!
!! RVEC_BRACKET3 finds the interval containing or nearest a given value.
!
!
!  Discussion:
!
!    The routine always returns the index LEFT of the sorted array
!    T with the property that either
!    *  T is contained in the interval [ T(LEFT), T(LEFT+1) ], or
!    *  T < T(LEFT) = T(1), or
!    *  T > T(LEFT+1) = T(N).
!
!    The routine is useful for interpolation problems, where
!    the abscissa must be located within an interval of data
!    abscissas for interpolation, or the "nearest" interval
!    to the (extreme) abscissa must be found so that extrapolation
!    can be carried out.
!
!  Modified:
!
!    05 April 1999
!
!  Author:
!
!    John Burkardt
!
!  Parameters:
!
!    Input, integer N, length of the input array.
!
!    Input, real T(N), an array sorted into ascending order.
!
!    Input, real TVAL, a value to be bracketed by entries of T.
!
!    Input/output, integer LEFT.
!
!    On input, if 1 <= LEFT <= N-1, LEFT is taken as a suggestion for the
!    interval [ T(LEFT), T(LEFT+1) ] in which TVAL lies.  This interval
!    is searched first, followed by the appropriate interval to the left
!    or right.  After that, a binary search is used.
!
!    On output, LEFT is set so that the interval [ T(LEFT), T(LEFT+1) ]
!    is the closest to TVAL; it either contains TVAL, or else TVAL
!    lies outside the interval [ T(1), T(N) ].
!
  implicit none
!
  integer n
!
  integer high
  integer left
  integer low
  integer mid
  real(8) t(n)
  real(8) tval
!
!  Check the input data.
!
  if ( n < 2 ) then
    write ( *, '(a)' ) ' '
    write ( *, '(a)' ) 'RVEC_BRACKET3 - Fatal error!'
    write ( *, '(a)' ) '  N must be at least 2.'
    stop
  end if
!
!  If LEFT is not between 1 and N-1, set it to the middle value.
!
  if ( left < 1 .or. left > n - 1 ) then
    left = ( n + 1 ) / 2
  end if
!
!  CASE 1: TVAL < T(LEFT):
!  Search for TVAL in [T(I), T(I+1)] for intervals I = 1 to LEFT-1.
!
  if ( tval < t(left) ) then

    if ( left == 1 ) then
      return
    else if ( left == 2 ) then
      left = 1
      return
    else if ( tval >= t(left-1) ) then
      left = left - 1
      return
    else if ( tval <= t(2) ) then
      left = 1
      return
    end if
!
!  ...Binary search for TVAL in [T(I), T(I+1)] for intervals I = 2 to LEFT-2.
!
    low = 2
    high = left - 2

    do

      if ( low == high ) then
        left = low
        return
      end if

      mid = ( low + high + 1 ) / 2

      if ( tval >= t(mid) ) then
        low = mid
      else
        high = mid - 1
      end if

    end do
!
!  CASE2: T(LEFT+1) < TVAL:
!  Search for TVAL in {T(I),T(I+1)] for intervals I = LEFT+1 to N-1.
!
  else if ( tval > t(left+1) ) then

    if ( left == n - 1 ) then
      return
    else if ( left == n - 2 ) then
      left = left + 1
      return
    else if ( tval <= t(left+2) ) then
      left = left + 1
      return
    else if ( tval >= t(n-1) ) then
      left = n - 1
      return
    end if
!
!  ...Binary search for TVAL in [T(I), T(I+1)] for intervals I = LEFT+2 to N-2.
!
    low = left + 2
    high = n - 2

    do

      if ( low == high ) then
        left = low
        return
      end if

      mid = ( low + high + 1 ) / 2

      if ( tval >= t(mid) ) then
        low = mid
      else
        high = mid - 1
      end if

    end do
!
!  CASE3: T(LEFT) <= TVAL <= T(LEFT+1):
!  T is in [T(LEFT), T(LEFT+1)], as the user said it might be.
!
  else

  end if

  return
end subroutine

!~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
end module